#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#


__docformat__ = 'restructuredtext'


# import standard library
import logging
logger = logging.getLogger('PyAMS (content.es)')

from multiprocessing import Process
from pprint import pformat
from threading import Thread

# import interfaces
from pyams_utils.interfaces import PYAMS_APPLICATION_SETTINGS_KEY, PYAMS_APPLICATION_DEFAULT_NAME
from transaction.interfaces import ITransactionManager
from zope.intid.interfaces import IIntIds

# import packages
from pyams_utils.registry import set_local_registry, get_utility
from pyams_utils.zodb import ZODBConnection
from pyams_zmq.handler import ZMQMessageHandler
from pyams_zmq.process import ZMQProcess
from pyramid.threadlocal import manager as threadlocal_manager
from zope.component.globalregistry import getGlobalSiteManager


class BaseIndexerProcess(Process):
    """Base indexer process"""

    def __init__(self, settings, group=None, target=None, name=None, *args, **kwargs):
        Process.__init__(self, group, target, name, *args, **kwargs)
        self.settings = settings

    def run(self):
        logger.debug("Starting indexer thread...")
        # Loading components registry
        registry = getGlobalSiteManager()
        threadlocal_manager.set({'request': None, 'registry': registry})
        logger.debug("Getting global registry: {0!r}".format(registry))
        # Get ES client
        es_client = getattr(registry, 'pyramid_es_client', None)
        if es_client is None:
            logger.debug("Missing ElasticSearch client in registry!")
            return
        # Check settings
        settings = self.settings
        logger.debug("Checking index parameters: {0}".format(str(settings)))
        document_id = settings.get('document')
        if not document_id:
            logger.warning('Bad indexer request: {0}'.format(str(settings)))
            return
        # Open ZODB connection
        manager = None
        zodb_name = settings.get('zodb_name')
        logger.debug("Opening ZODB connection...")
        zodb = ZODBConnection(name=zodb_name)
        connection = zodb.get_connection()
        try:
            root = connection.root()
            logger.debug("Getting connection root {0!r}".format(root))
            application_name = registry.settings.get(PYAMS_APPLICATION_SETTINGS_KEY, PYAMS_APPLICATION_DEFAULT_NAME)
            application = root.get(application_name)
            logger.debug("Loading application {0!r} named {1}".format(application, application_name))
            if application is not None:
                # set local registry
                sm = application.getSiteManager()
                set_local_registry(sm)
                logger.debug("Setting local registry {0!r}".format(sm))
                # find document
                intids = get_utility(IIntIds)
                document = intids.queryObject(document_id)
                if document is None:
                    logger.warning("Can't find requested document {0}!".format(document_id))
                    return
                # index document
                logger.debug("Starting indexing for {0!r}".format(document))
                manager = ITransactionManager(document)
                for attempt in manager.attempts():
                    with attempt as t:
                        self.update_index(es_client, document)
                    if t.status == 'Committed':
                        break
        finally:
            if manager is not None:
                manager.abort()
            connection.close()
            threadlocal_manager.pop()

    def update_index(self, client, document):
        """Update index"""
        raise NotImplementedError("Index threads must implement update_index method")


class IndexerProcess(BaseIndexerProcess):
    """Content indexer process"""

    def update_index(self, client, document):
        client.index_object(document)


class UnindexerProcess(BaseIndexerProcess):
    """Content un-indexer process"""

    def update_index(self, client, document):
        client.delete_object(document)


class IndexerThread(Thread):
    """Content indexer thread"""

    def __init__(self, process):
        Thread.__init__(self)
        self.process = process

    def run(self):
        self.process.start()
        self.process.join()


class ContentIndexerHandler(object):
    """Content indexer handler"""

    def index(self, settings):
        IndexerThread(IndexerProcess(settings)).start()
        return [200, 'Indexer process started']

    def unindex(self, settings):
        IndexerThread(UnindexerProcess(settings)).start()
        return [200, 'Un-indexer process started']

    def test(self, settings):
        messages = ['Content indexer process ready to handle requests.', '']
        registry = getGlobalSiteManager()
        es_client = getattr(registry, 'pyramid_es_client', None)
        if es_client is None:
            messages.append('WARNING: no ElasticSearch client defined!')
        else:
            messages.extend(['ElasticSearch client properties:',
                             '- servers: {0}'.format(es_client.es.transport.hosts),
                             '- index: {0}'.format(es_client.index),
                             ''])
            ping = es_client.es.ping()
            messages.append('Server ping: {0}'.format('OK' if ping else 'KO'))
            if ping:
                messages.extend(['', 'Server info:'])
                messages.extend(pformat(es_client.es.info()).split('\n'))
        return [200, '\n'.join(messages)]


class ContentIndexerMessageHandler(ZMQMessageHandler):
    """Content indexer message handler"""

    handler = ContentIndexerHandler


class ContentIndexerProcess(ZMQProcess):
    """Content indexer ZMQ process"""

    def __init__(self, zmq_address, handler, registry):
        ZMQProcess.__init__(self, zmq_address, handler)
        self.registry = registry
